"""Packages"""
import numpy.random as rnd
import numpy, scipy, pylab, math, time, sys, matplotlib.pyplot as plt 
from scipy import integrate
from scipy import fftpack
#import vegas
import cmath
import itertools
from multiprocessing import Pool
import os
"""Set timer"""
start_time = time.time()
"""Input Parameters"""
electron_temperature = 12.0 ; """electron temperature in eV""" 
electron_density = 1.808e29 ;"""electron density in m^-3""" 
k_vector = 0.378e10; """k vector in m^-1 """
"""Constants in atomic units"""
h=1.0; """Planck constant"""
m=0.5; """electron mass""" 
e=math.sqrt(2); """electron charge"""
eps=0.25/(math.pi); """permitivity of vacuum""" 
a_b=1.0; """Bohr radius """
K_b=b=1.0; """Boltzmann constant"""
w=numpy.linspace(-5,5,500); """Freqeuncy range"""
w_in = numpy.linspace(-25,25,1000); """ Frequency grid for output  """
#p_in = 1.0 ;
p_points = numpy.linspace(-3,3,1000); """ For integration of q-points """
"""Plasma Parameters"""
T = Tm = electron_temperature*1.0/13.6 ; """electron temperature"""
n_e = (electron_density)*(10**-30)*(0.5291**3); """electron density""" 
q = k = k_vector*(0.5291)/(10**10); """k-vector"""
T_f = t = 1.0*(h*h/(2*m*K_b))*( ( 3*math.pi*math.pi*n_e )**((2.0/3))); """Fermi temperature"""
k_f = (2.0*m*K_b*T_f/(h*h))**(0.5);"""Fermi K-vector"""   #output in 1/a.u.
w_pl = (n_e*4*math.pi*e*e/(m))**(0.5);"""Plasma frequency"""   #output in Rydberg  
Lambda_D = ( n_e / Tm )**0.5 ; """Debye inverse screening length"""
w_q =  ( (w_pl * ( 1.0 + (q/k)**2 ) ) + (q**4) )**0.5 ; """Plasmon satellite frequency"""
e_coupling = ( (4.0*numpy.pi*n_e/3.0)**(0.333) )/( 0.5*Tm);
e_degeneracy = Tm / (q)**2 ;
parameters={ 'electron_temperature': Tm,'electron_density': n_e, 'plasma_frequency':w_pl, 'electron_degeneracy' : e_degeneracy , 'electron_coupling' : e_coupling } 

#######################################################################################################  
"""Chemical potential"""
cpcheck2 = 2.0*math.pi*math.pi*n_e; 
errorcheck=[];
chempot=numpy.linspace(-100*Tm,100*Tm,50000);  
#Loop to check the right chemical potential
def cp(p):
  return 2.0*(p**2)/(1.0 + scipy.exp( ( (p**2) - chempot[i])/(K_b*Tm) ) ); 
"""Find the convergence by minimizing the difference between numerical and analytical solution"""
for i in range(len(chempot)):
  numans = scipy.integrate.quad(cp,0,numpy.inf)[0]
  errorcheck.append( abs(numans - cpcheck2) )
  #if abs(numans - cpcheck2) >= 10**(-03)  and abs(numans - cpcheck2) <= 10**(-02)  :
  #print "error = %f  chemical potential = %f " %(abs(numans - cpcheck2), chempot[i])
#gives the index of least error:  errorcheck.index(min(errorcheck))
g = gamma = mu = chempot[ errorcheck.index(min(errorcheck))  ];
##############################################################################################################################
imvalues=[];
re_rpa=[];
"""Imaginary part of dielectric function as a function of frequency"""
def imrpa(w):
	return (2.0*m*e*e*h*h/(q**3))*(1.0*T*m*scipy.log(1.0*scipy.exp((-g + 0.5*(-1.0*h*m*w/q + 0.5*q)**2/m)/T) + 1.0) - 1.0*T*m*scipy.log(1.0*scipy.exp((-g + 0.5*(1.0*h*m*w/q + 0.5*q)**2/m)/T) + 1.0) - 0.5*(-1.0*h*m*w/q + 0.5*q)**2 + 0.5*(1.0*h*m*w/q + 0.5*q)**2)

"""Real part of dielectric function as a function of momentum"""
def re_rpa_fn(p):
	return (4.0*0.5*m*e*e/(math.pi*h*q*q*q))*( scipy.log( (p*a_b/h) - (0.5*q*a_b/h) - (m*w[i]*a_b/q) ) + scipy.log( (p*a_b/h) - (0.5*q*a_b/h) + (m*w[i]*a_b/q) ) )*p*(1.0/(1.0 + scipy.exp( ((0.5*p*p/m) - gamma)/(K_b*T))))
	
"""Imaginary part of dielectric function at given frequencies"""
#imvalues = map( imrpa, w );
"""Real part of dielectric function at given frequencies"""
#for i in range(len(w)):
#  re_rpa.append( 1.0 - scipy.integrate.quad(re_rpa_fn, -numpy.inf,  numpy.inf)[0] )
########################################################################################################################
"""imaginary part of dielectric function as a function of frequency and momentum"""
imrpa2 = lambda q_grid, w : (2.0*m*e*e*h*h/(q_grid**3))*(1.0*T*m*scipy.log(1.0*scipy.exp((-g + 0.5*(-1.0*h*m*w/q_grid + 0.5*q_grid  )**2/m )/T) + 1.0) - 1.0*T*m*scipy.log(1.0*scipy.exp((-g + 0.5*(1.0*h*m*w/q_grid + 0.5*q_grid)**2/m)/T)  + 1.0) - 0.5*(-1.0*h*m*w/q_grid + 0.5*q_grid)**2 + 0.5*(1.0*h*m*w/q_grid + 0.5*q_grid)**2)

"""Real part of dielectric function as a function of momentum"""
re_rpa_grid = lambda p : (4.0*0.5*m*e*e/(math.pi*h*q_grid[j]*q_grid[j]*q_grid[j]))*( scipy.log( (p*a_b/h) - (0.5*q_grid[j]*a_b/h) - (m*w[i]*a_b/q_grid[j]) ) + scipy.log( (p*a_b/h) - (0.5*q_grid[j]*a_b/h) + (m*w[i]*a_b/q_grid[j]) ) )*p*(1.0/(1.0 + scipy.exp( ((0.5*p*p/m) - gamma)/(K_b*T))))
	

########################################################################################################################
"""Imaginary part of rpa dielectric function as a function of frequency and momentum"""
def im_rpa_pw( p, w ) :
 return (2.0*m*e*e*h*h/(p**3))*(1.0*T*m*numpy.log(1.0*numpy.exp((-g + 0.5*(-1.0*h*m*w/p + 0.5*p)**2/m)/T) + 1.0) - 1.0*T*m*numpy.log(1.0*numpy.exp((-g + 0.5*(1.0*h*m*w/p + 0.5*p)**2/m)/T) + 1.0) - 0.5*(-1.0*h*m*w/p + 0.5*p)**2 + 0.5*(1.0*h*m*w/p + 0.5*p)**2)
"""Real part of rpa dielectric function as a function of frequency and momentum in the right form which is passed after integration """
def re_rpa_pw(p, w ) : 
    return 1.0 - scipy.integrate.quad( lambda y : (4.0*0.5*m*e*e/(scipy.pi*h*p*p*p))*( scipy.log( (y*a_b/h) - (0.5*p*a_b/h) - (m*w*a_b/p) ) + scipy.log( (y*a_b/h) - (0.5*p*a_b/h) + (m*w*a_b/p) ) )*y*(1.0/(1.0 + scipy.exp( ((0.5*y*y/m) - gamma)/(K_b*T))))	, -100, 100 )[0] 

########################################################################################################################
"""Imaginary part of the inverse rpa dielectric function as a function of frequency and momentum"""
inv_im = lambda eps1, eps2:  eps2/( eps1**2 + eps2**2  )   
########################################################################################################################
############################################################################################################
#Kramers-Kronig for im-rpa function to obtain re-rpa using principal value Cauchy integration
def re_rpa_fn_kk(p,w): #Returns Re-RPA function
 q=p;   
 def re(w):
   return  (1.0/scipy.pi) *( (1.0/(w - y[i] ))**0 )*(2.0*m*e*e*h*h/(q**3))*(1.0*T*m*scipy.log(1.0*scipy.exp((-g + 0.5*(-1.0*h*m*w/q + 0.5*q)**2/m)/T) + 1.0) - 1.0*T*m*scipy.log(1.0*scipy.exp((-g + 0.5*(1.0*h*m*w/q + 0.5*q)**2/m)/T) + 1.0) - 0.5*(-1.0*h*m*w/q + 0.5*q)**2 + 0.5*(1.0*h*m*w/q + 0.5*q)**2)

 cc=[];
 y=numpy.array([w]);
 for i in range(len(y)):
  cc.append( scipy.integrate.quad(re, -20,20,weight='cauchy', wvar=y[i] )[0] + 1.0 )
 return cc[0]
############################################################################################
#Kramers-Kronig for re-rpa function to obtain im-rpa using principal value Cauchy integration
def im_rpa_fn_kk(p,w): #Returns Im-RPA function
 #q=p;   
 def im(w):
   return  -(1.0/scipy.pi) *( (1.0/(w - y[i] ))**0 ) * ( 1.0 -  scipy.integrate.quad( lambda z : (4.0*0.5*m*e*e/(math.pi*h*p*p*p))*( scipy.log( (z*a_b/h) - (0.5*p*a_b/h) - (m*w*a_b/p) ) + scipy.log( (z*a_b/h) - (0.5*p*a_b/h) + (m*w*a_b/p) ) )*z*(1.0/(1.0 + scipy.exp( ((0.5*z*z/m) - gamma)/(K_b*T))))	, -20, 20 )[0] )

 cc=[];
 y=numpy.array([w]);
 for i in range(len(y)):
  cc.append( scipy.integrate.quad(im, -10,10,weight='cauchy', wvar=y[i] )[0]  )
 return cc[0] 
############################################################################################
inv_im = lambda eps1, eps2: - eps2/( eps1**2 + eps2**2  )
import numpy as np 
freq_range=np.linspace(0,8.0,700)   

#q in im_rpa_pw(q,w) and re_rpa_pw(q,w) below is in a.u. 
im_values=[]
re_values=[]
im_inv_eps_values=[] 
for i in range(len(freq_range)): 
 im_values.append(im_rpa_pw(1.743, freq_range[i]))
 re_values.append(re_rpa_pw(1.743, freq_range[i]))
 im_inv_eps_values.append( im_values[i] / ( (im_values[i])**2 + (re_values[i])**2  )  ) 

#########define DSF#########
#import numpy as np
#s_q_w = lambda eps1, eps2, q, T, w, ne : (np.pi**0) * (ne**-1)  * (q**2) * (eps2/( eps1**2 + eps2**2  )) / ( 1.0 - np.exp(-w/T) )
#Ex: s_q_w(re_eps_1,im_eps_1, 0.5773, 13, freq_1_ev, 0.109 )
##all in a.u.
#dsf=[]
#for i in range(len(freq_range)):
# dsf.append( s_q_w( re_values[i], im_values[i], 0.23, 12, freq_range[i], 0.17356 )   )  

#########define DSF#########
import numpy as np
s_q_w = lambda eps1, eps2, q, T, w, ne : (np.pi**0) * (ne**-1)  * (q**2) * (eps2/( eps1**2 + eps2**2  )) / ( 1.0 - np.exp(-w/T) )
#Ex: s_q_w(re_eps_1,im_eps_1, 0.5773, 13, freq_1_ev, 0.109 )
##all in a.u.
dsf=[]
for i in range(len(freq_range)):       
 dsf.append( s_q_w( re_values[i], im_values[i], 1.743, 12, freq_range[i], 0.17356 )   )  


#######write freq, Im(\epsilon^{-1}), dsf, LFC, T-LFC to a file##########
file = open("al_rs_2p07_12ev_qoverqf_1.88.dat", "w")  
#print("#freq (a.u.)        im[eps^-1]          S(q,w) (a.u.) ")   
for index in range(len(freq_range)):
    file.write(  str(freq_range[index]/w_pl) + " " + str(im_inv_eps_values[index]*1.0)  +  " " +  str(dsf[index]*1.0) +   "\n")
file.close()     
##############Plots#################################
print("plasma frequency: ",  w_pl) 
#plt.plot( freq_range, im_values ) 
#plt.plot( freq_range, re_values )
plt.plot( freq_range/w_pl, im_inv_eps_values, 'k--', label=r'$ Im(\epsilon^{-1}) $' )
plt.plot( freq_range/w_pl, im_values, 'r:', label=r'$ Im(\epsilon) $'  )
plt.plot( freq_range/w_pl, re_values, 'b-.,', label=r'$ Re(\epsilon) $'  )
plt.plot( freq_range/w_pl, dsf, 'g', label=r'$ S(q,\omega) $'  )
#plt.xlim(0,2)  
#plt.ylim(-2,10) 
plt.legend(loc='upper right',fontsize=10,frameon=False, ncol=1, numpoints=1, scatterpoints = 1 )
plt.xlabel( r'$ \frac{ \omega  }{ \omega_{pl}  }  $' , fontsize=15  )
plt.ylabel( r'$ \epsilon [q,\omega] $' , fontsize=15  )   
plt.show()
    
 

#################################################

##Ignore
#freq_ry = numpy.linspace(0,40,10)  
#freq_ev = freq_ry * 27.211 / 2.0
#mom_vec = 0.5   
#im_val=[]  
#re_val=[] 
#inv_im_val=[]
#for i in range(len(freq_ry)):
# im_val.append( im_rpa_pw(mom_vec,freq_ry[i]) )
# re_val.append( re_rpa_pw(mom_vec,freq_ry[i]) )

#for i in range(len(im_val)):
# inv_im_val.append( inv_im(re_val[i],im_val[i])  )

#print(im_val)
#plt.plot( freq_ev, inv_im_val  )
#plt.xlabel(r'$\omega \ [eV]$',fontsize=15)  
#plt.ylabel(r'$\Im[\epsilon^{-1}(q,\omega)]$',fontsize=15) 
#plt.show()    
#####


